#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#


from typing import Any, Mapping

from airbyte_cdk import emit_configuration_as_airbyte_control_message
from airbyte_cdk.sources.file_based.file_based_source import FileBasedSource
from source_gcs.legacy_config_transformer import LegacyConfigTransformer
from source_gcs.spec import SourceGCSSpec


class SourceGCS(FileBasedSource):
    @classmethod
    def read_config(cls, config_path: str) -> Mapping[str, Any]:
        """
        Override the default read_config to transform the legacy config format
        into the new one before validating it against the new spec.
        """
        config = FileBasedSource.read_config(config_path)
        if not cls._is_file_based_config(config):
            parsed_legacy_config = SourceGCSSpec(**config)
            converted_config = LegacyConfigTransformer.convert(parsed_legacy_config)
            emit_configuration_as_airbyte_control_message(converted_config)
            return converted_config
        return config

    @staticmethod
    def _is_file_based_config(config: Mapping[str, Any]) -> bool:
        return "streams" in config



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#


from airbyte_cdk.sources.file_based.config.abstract_file_based_spec import AbstractFileBasedSpec
from pydantic.v1 import AnyUrl, Field


class Config(AbstractFileBasedSpec):
    """
    NOTE: When this Spec is changed, legacy_config_transformer.py must also be
    modified to uptake the changes because it is responsible for converting
    legacy GCS configs into file based configs using the File-Based CDK.
    """

    service_account: str = Field(
        title="Service Account Information",
        airbyte_secret=True,
        description=(
            "Enter your Google Cloud "
            '<a href="https://cloud.google.com/iam/docs/creating-managing-service-account-keys#creating_service_account_keys">'
            "service account key</a> in JSON format"
        ),
        order=0,
    )

    bucket: str = Field(title="Bucket", description="Name of the GCS bucket where the file(s) exist.", order=2)

    @classmethod
    def documentation_url(cls) -> AnyUrl:
        """
        Returns the documentation URL.
        """
        return AnyUrl("https://docs.airbyte.com/integrations/sources/gcs", scheme="https")

    @staticmethod
    def remove_discriminator(schema) -> None:
        pass



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#


from pydantic import BaseModel, Field


class SourceGCSSpec(BaseModel):
    """
    The SourceGCSSpec class defines the expected input configuration
    for the Google Cloud Storage (GCS) source. It uses Pydantic for data
    validation through the defined data models.

    Note: When this Spec is changed, ensure that the legacy_config_transformer.py
    is also modified to accommodate the changes, as it is responsible for
    converting legacy GCS configs into file based configs using the File-Based CDK.
    """

    gcs_bucket: str = Field(
        title="GCS bucket",
        description="GCS bucket name",
        order=0,
    )

    gcs_path: str = Field(
        title="GCS Path",
        description="GCS path to data",
        order=1,
    )

    service_account: str = Field(
        title="Service Account Information.",
        airbyte_secret=True,
        description=(
            'Enter your Google Cloud <a href="https://cloud.google.com/iam/docs/'
            'creating-managing-service-account-keys#creating_service_account_keys">'
            "service account key</a> in JSON format"
        ),
    )



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#


import json

from google.cloud import storage
from google.oauth2 import service_account


def get_gcs_client(config):
    credentials = service_account.Credentials.from_service_account_info(json.loads(config.service_account))
    client = storage.Client(credentials=credentials)
    return client


def get_gcs_blobs(config):
    client = get_gcs_client(config)
    bucket = client.get_bucket(config.gcs_bucket)
    blobs = bucket.list_blobs(prefix=config.gcs_path)
    # TODO: only support CSV initially. Change this check if implementing other file formats.
    blobs = [blob for blob in blobs if "csv" in blob.name.lower()]
    return blobs


def get_stream_name(blob):
    blob_name = blob.name
    # Remove path from stream name
    blob_name_without_path = blob_name.split("/")[-1]
    # Remove file extension from stream name
    stream_name = blob_name_without_path.replace(".csv", "")
    return stream_name



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#


import logging
from datetime import datetime

from airbyte_cdk.sources.file_based.remote_file import RemoteFile
from airbyte_cdk.sources.file_based.stream.cursor import DefaultFileBasedCursor


class Cursor(DefaultFileBasedCursor):
    @staticmethod
    def get_file_uri(file: RemoteFile) -> str:
        return file.uri.split("?")[0]

    def add_file(self, file: RemoteFile) -> None:
        uri = self.get_file_uri(file)
        self._file_to_datetime_history[uri] = file.last_modified.strftime(self.DATE_TIME_FORMAT)
        if len(self._file_to_datetime_history) > self.DEFAULT_MAX_HISTORY_SIZE:
            # Get the earliest file based on its last modified date and its uri
            oldest_file = self._compute_earliest_file_in_history()
            if oldest_file:
                del self._file_to_datetime_history[oldest_file.uri]
            else:
                raise Exception(
                    "The history is full but there is no files in the history. This should never happen and might be indicative of a bug in the CDK."
                )

    def _should_sync_file(self, file: RemoteFile, logger: logging.Logger) -> bool:
        uri = self.get_file_uri(file)
        if uri in self._file_to_datetime_history:
            # If the file's uri is in the history, we should sync the file if it has been modified since it was synced
            updated_at_from_history = datetime.strptime(self._file_to_datetime_history[uri], self.DATE_TIME_FORMAT)
            if file.last_modified < updated_at_from_history:
                logger.warning(
                    f"The file {uri}'s last modified date is older than the last time it was synced. This is unexpected. Skipping the file."
                )
            else:
                return file.last_modified > updated_at_from_history
            return file.last_modified > updated_at_from_history
        if self._is_history_full():
            if self._initial_earliest_file_in_history is None:
                return True
            if file.last_modified > self._initial_earliest_file_in_history.last_modified:
                # If the history is partial and the file's datetime is strictly greater than the earliest file in the history,
                # we should sync it
                return True
            elif file.last_modified == self._initial_earliest_file_in_history.last_modified:
                # If the history is partial and the file's datetime is equal to the earliest file in the history,
                # we should sync it if its uri is strictly greater than the earliest file in the history
                return uri > self._initial_earliest_file_in_history.uri
            else:
                # Otherwise, only sync the file if it has been modified since the start of the time window
                return file.last_modified >= self.get_start_time()
        else:
            # The file is not in the history and the history is complete. We know we need to sync the file
            return True



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#


import sys
import traceback
from datetime import datetime

from airbyte_cdk import AirbyteEntrypoint, AirbyteMessage, Type, launch
from airbyte_cdk.models import AirbyteErrorTraceMessage, AirbyteTraceMessage, TraceType
from source_gcs import Config, Cursor, SourceGCS, SourceGCSStreamReader


def run():
    _args = sys.argv[1:]
    try:
        catalog_path = AirbyteEntrypoint.extract_catalog(_args)
        config_path = AirbyteEntrypoint.extract_config(_args)
        state_path = AirbyteEntrypoint.extract_state(_args)
        source = SourceGCS(
            SourceGCSStreamReader(),
            Config,
            SourceGCS.read_catalog(catalog_path) if catalog_path else None,
            SourceGCS.read_config(config_path) if config_path else None,
            SourceGCS.read_state(state_path) if state_path else None,
            cursor_cls=Cursor,
        )
    except Exception:
        print(
            AirbyteMessage(
                type=Type.TRACE,
                trace=AirbyteTraceMessage(
                    type=TraceType.ERROR,
                    emitted_at=int(datetime.now().timestamp() * 1000),
                    error=AirbyteErrorTraceMessage(
                        message="Error starting the sync. This could be due to an invalid configuration or catalog. Please contact Support for assistance.",
                        stack_trace=traceback.format_exc(),
                    ),
                ),
            ).json()
        )
    else:
        launch(source, sys.argv[1:])



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#


import itertools
import json
import logging
from datetime import datetime, timedelta
from io import IOBase
from typing import Iterable, List, Optional

import pytz
import smart_open
from airbyte_cdk.sources.file_based.exceptions import ErrorListingFiles, FileBasedSourceError
from airbyte_cdk.sources.file_based.file_based_stream_reader import AbstractFileBasedStreamReader, FileReadMode
from airbyte_cdk.sources.file_based.remote_file import RemoteFile
from google.cloud import storage
from google.oauth2 import service_account
from source_gcs.config import Config

ERROR_MESSAGE_ACCESS = (
    "We don't have access to {uri}. The file appears to have become unreachable during sync."
    "Check whether key {uri} exists in `{bucket}` bucket and/or has proper ACL permissions"
)


class SourceGCSStreamReader(AbstractFileBasedStreamReader):
    """
    Stream reader for Google Cloud Storage (GCS).
    """

    def __init__(self):
        super().__init__()
        self._gcs_client = None
        self._config = None

    @property
    def config(self) -> Config:
        return self._config

    @config.setter
    def config(self, value: Config):
        assert isinstance(value, Config), "Config must be an instance of the expected Config class."
        self._config = value

    def _initialize_gcs_client(self):
        if self.config is None:
            raise ValueError("Source config is missing; cannot create the GCS client.")
        if self._gcs_client is None:
            credentials = self._get_credentials()
            self._gcs_client = storage.Client(credentials=credentials)
        return self._gcs_client

    def _get_credentials(self):
        return service_account.Credentials.from_service_account_info(json.loads(self.config.service_account))

    @property
    def gcs_client(self) -> storage.Client:
        return self._initialize_gcs_client()

    def get_matching_files(self, globs: List[str], prefix: Optional[str], logger: logging.Logger) -> Iterable[RemoteFile]:
        """
        Retrieve all files matching the specified glob patterns in GCS.
        """
        try:
            start_date = (
                datetime.strptime(self.config.start_date, self.DATE_TIME_FORMAT) if self.config and self.config.start_date else None
            )
            prefixes = [prefix] if prefix else self.get_prefixes_from_globs(globs or [])
            globs = globs or [None]

            for prefix, glob in itertools.product(prefixes, globs):
                bucket = self.gcs_client.get_bucket(self.config.bucket)
                blobs = bucket.list_blobs(prefix=prefix, match_glob=glob)
                for blob in blobs:
                    last_modified = blob.updated.astimezone(pytz.utc).replace(tzinfo=None)

                    if not start_date or last_modified >= start_date:
                        uri = blob.generate_signed_url(expiration=timedelta(hours=1), version="v4")

                        file_extension = ".".join(blob.name.split(".")[1:])

                        yield RemoteFile(uri=uri, last_modified=last_modified, mime_type=file_extension)

        except Exception as exc:
            self._handle_file_listing_error(exc, prefix, logger)

    def _handle_file_listing_error(self, exc: Exception, prefix: str, logger: logging.Logger):
        logger.error(f"Error while listing files: {str(exc)}")
        raise ErrorListingFiles(
            FileBasedSourceError.ERROR_LISTING_FILES,
            source="gcs",
            bucket=self.config.bucket,
            prefix=prefix,
        ) from exc

    def open_file(self, file: RemoteFile, mode: FileReadMode, encoding: Optional[str], logger: logging.Logger) -> IOBase:
        """
        Open and yield a remote file from GCS for reading.
        """
        logger.debug(f"Trying to open {file.uri}")

        # choose correct compression mode
        file_extension = file.mime_type.split(".")[-1]
        if file_extension in ["gz", "bz2"]:
            compression = "." + file_extension
        else:
            compression = "disable"

        try:
            result = smart_open.open(file.uri, mode=mode.value, compression=compression, encoding=encoding)
        except OSError as oe:
            logger.warning(ERROR_MESSAGE_ACCESS.format(uri=file.uri, bucket=self.config.bucket))
            logger.exception(oe)
        return result



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#

from .config import Config
from .cursor import Cursor
from .legacy_config_transformer import LegacyConfigTransformer
from .source import SourceGCS
from .stream_reader import SourceGCSStreamReader

__all__ = [
    "Config",
    "Cursor",
    "LegacyConfigTransformer",
    "SourceGCS",
    "SourceGCSStreamReader",
]



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#


from typing import Any, Dict, Mapping

from source_gcs.spec import SourceGCSSpec

from .helpers import get_gcs_blobs, get_stream_name


class LegacyConfigTransformer:
    """
    Transforms GCS source configs from legacy format to be compatible
    with the new GCS source built with the file-based CDK.
    """

    @staticmethod
    def _create_stream(blob: Any, legacy_prefix: str) -> Dict[str, Any]:
        """
        Create a stream dict from a blob.

        :param blob: The blob from which to create the stream.
        :param legacy_prefix: The legacy prefix path on GCS.
        :return: A dictionary representing the stream.
        """
        return {
            "name": get_stream_name(blob),
            "legacy_prefix": f"{legacy_prefix}/{blob.name.split('/')[-1]}",
            "validation_policy": "Emit Record",
            "format": {"filetype": "csv"},
        }

    @classmethod
    def convert(cls, legacy_config: SourceGCSSpec) -> Mapping[str, Any]:
        """
        Convert a legacy configuration to a transformed configuration.

        :param legacy_config: Legacy configuration of type SourceGCSSpec.
        :return: Transformed configuration as a dictionary.
        """
        blobs = get_gcs_blobs(legacy_config)
        streams = [cls._create_stream(blob, legacy_config.gcs_path) for blob in blobs]

        return {"bucket": legacy_config.gcs_bucket, "service_account": legacy_config.service_account, "streams": streams}
